"""
open command
"""

import logging
import time

from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.firefox.options import Options
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from asrch.modules._formatting import Bar
from asrch.modules.logging_formatter import ColorFormatter



form = Bar()
options = Options()

c_log = logging.getLogger(__name__)
c_log.setLevel(logging.DEBUG)

sh = logging.StreamHandler()

c_form = ColorFormatter("%(asctime)s|%(levelname)8s|%(message)s")

sh.setFormatter(c_form)

c_log.addHandler(sh)


def get_quizlet(self, url: str, proxy: str, header: bool):
    """Gets the content of the <body> tag from a specified URL.

    :param url: The URL from which to fetch the <body> content.
    :type url: str
    :param proxy: The proxy server address to use for the request.
    :type proxy: str
    :param header: Whether to include custom headers in the request.
    :type header: bool
    :return: A list of strings containing the <body> content.
    :rtype: list
    """

    c_log.debug(self.url)
    c_log.debug(self.proxy)
    c_log.debug(self.header)

    if header:
        c_log.info("Headless false")
    else:
        c_log.info("Headless true")
        options.add_argument("--headless")

    if proxy is not None:
        c_log.info("Proxy set")
    elif proxy is None:
        c_log.debug("no proxy set")

    with webdriver.Firefox(options=options) as driver:
        output: list[str] = []
        c_log.info("Getting quizlet set")
        driver.get(url)
        time.sleep(3)
        WebDriverWait(driver, 4)
        body = WebDriverWait(driver, 10).until(
            EC.visibility_of_element_located((By.TAG_NAME, "body"))
        )

        output.append(body.text)
        return "\n".join(output)



"""
Constants
"""

from enum import Enum
from pathlib import Path


class OpenModes(Enum):
    """
    Modes for the open command
    """

    TEXT = "text"
    BODY = "body"
    JS = "js"
    HTML = "html"
    IMG = "img"
    SS = "ss"
    INS = "inspect"


class Colors:
    # Main stuff
    GREEN = "\x1b[38;5;40m"
    RED = "\x1b[38;5;196m"
    ORANGE = "\x1b[38;5;202m"
    YELLOW = "\x1b[38;5;226m"
    BLUE = "\x1b[38;5;21m"
    NC = "\x1b[0m"
    GRAY = "\x1b[38;5;243m"
    MAGENTA = "\x1b[45m"
    DIM_GRAY = "\x1b[2;23m"
    BLACK = "\033[30m"

    # Special
    CRITICAL = "\x1b[30;41m"
    TIME = "\x1b[35;3m"

    BOLD = "\033[1m"
    NOBOLD = "\033[0m"

    # HREF
    HREF = "\x1b[2;23m"

    # Red shades
    RED_DARK = "\x1b[38;5;196m"
    RED_MEDIUM_DARK = "\x1b[38;5;202m"
    RED_MEDIUM = "\x1b[38;5;208m"
    RED_MEDIUM_LIGHT = "\x1b[38;5;214m"
    RED_LIGHT = "\x1b[38;5;220m"

    # Happy
    PINK = "\x1b[38;5;206m"
    PURPLE = "\x1b[38;5;141m"
    CYAN = "\x1b[38;5;87m"
    TEAL = "\x1b[38;5;37m"
    LILAC = "\x1b[38;5;183m"
    MINT = "\x1b[38;5;84m"

    # Pastel Colors
    PASTEL_PINK = "\x1b[38;5;218m"  
    PASTEL_PURPLE = "\x1b[38;5;183m"  
    PASTEL_BLUE = "\x1b[38;5;153m" 
    PASTEL_GREEN = "\x1b[38;5;120m" 
    PASTEL_YELLOW = "\x1b[38;5;226m"  
    PASTEL_ORANGE = "\x1b[38;5;223m"  
    PASTEL_CYAN = "\x1b[38;5;87m"  
    PASTEL_LILAC = "\x1b[38;5;183m"  
    PASTEL_MINT = "\x1b[38;5;121m"  

    # Background Colors
    BG_GRAY = "\x1b[48;5;238m" 
    BG_DARK_GRAY = "\x1b[48;5;236m" 
    BG_BLACK = "\x1b[48;5;232m" 
    BG_BLUE = "\x1b[48;5;27m" 
    BG_GREEN = "\x1b[48;5;28m"  
    BG_RED = "\x1b[48;5;52m" 
    BG_YELLOW = "\x1b[48;5;226m"  
    BG_ORANGE = "\x1b[48;5;202m"  
    BG_PURPLE = "\x1b[48;5;129m"  
    BG_PINK = "\x1b[48;5;206m"  
    BG_CYAN = "\x1b[48;5;45m"  
    BG_TEAL = "\x1b[48;5;30m"  


class Files:
    ROOT_DIR = Path(__file__).parent
    FILE1 = ROOT_DIR / "open.py"



"""
class ConfigDefaults:
    open_header = config['commands']['open']['header']
    open_pager = config['commands']['open']['pager']
    open_download = config['commands']['open']['download']
    open_log = config['commands']['open']['log']
    open_nodriver = config['commands']['open']['nodriver']
    open_parser = config['commands']['open']['parser']

    search_query = config['commands']['search']['query']
    search_header = config['commands']['search']['header']
    search_nodriver = config['commands']['search']['nodriver']
    search_log = config['commands']['search']['log']

    whois_header = config['commands']['whois']['header']

    find_element = config['commands']['find']['element']
    find_header = config['commands']['find']['header']
    find_log = config['commands']['find']['log']
    find_locator = config['commands']['find']['locator']

    quizlet_header = config['commands']['quizlet']['header']
    quizlet_pager = config['commands']['quizlet']['pager']
"""



"""
open command
"""

import logging
import os
import random
import shutil
import sys
import time
import urllib
import urllib.request
from collections.abc import Generator
from datetime import datetime
from typing import Optional

import requests
from ascli.modules._box import Box
from ascli.modules._output import Output
from ascli.modules._alert import Alert
from bs4 import BeautifulSoup
from selenium import webdriver
from selenium.common.exceptions import (
    ElementNotVisibleException,
    NoSuchElementException,
    StaleElementReferenceException,
)
from selenium.webdriver.common.by import By
from selenium.webdriver.firefox.options import Options
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from asrch.modules._formatting import Bar
from asrch.modules.logging_formatter import ColorFormatter
from asrch.utils.constants import Colors

today = datetime.today().date()
#formatted_date = today.strftime("%m%d")  
form = Bar()
options = Options()
colors = Colors()
history: list[str] = []  # session history for browse mode
terminal_size = round(shutil.get_terminal_size().columns)

current_dir = os.path.dirname(os.path.realpath(__file__))
output_file_path = os.path.join(current_dir, "../../../workspaces/")
resolved_path = os.path.abspath(output_file_path)

c_log = logging.getLogger(__name__)
sh = logging.StreamHandler()

c_form = ColorFormatter("%(asctime)s|%(levelname)8s|%(message)s")

sh.setFormatter(c_form)

c_log.addHandler(sh)


def highlight_elements(soup):
    """
    Highlight all elements in BeautifulSoup object
    """
    for tag in soup.find_all(["h1"]):
        if tag.string is not None:
            tag.string = Box.create_box(None, tag.string, None, padding=2, style="h1")

    for tag in soup.find_all(["h2"]):
        if tag.string is not None:
            tag.string = Box.create_box(None, tag.string, None, padding=2, style="h2")

    for tag in soup.find_all(["h3", "p"]):
        if tag.string is not None:
            tag.string = Box.create_box(None, tag.string, None, padding=2, style="h3")

    for tag in soup.find_all("a", href=True):
        href = tag.get("href")
        if href:
            tag.string = f"{tag.string}({href})"
    return soup


def get_page(
    url: str | int,
    header: bool = False,
    proxy: Optional[str] = None,
    log=True,
    parser="html.parser",
    browse: bool = False,
) -> str | int:
    """Get the content of a web page.

    :param url: The URL of the web page.
    :type url: str
    :param header: Flag indicating whether to include headers in the request, defaults to False.
    :type header: bool, optional
    :param proxy: Proxy to be used for the request, defaults to None.
    :type proxy: Optional[str], optional
    :param log: Flag indicating whether to log actions, defaults to True.
    :type log: bool, optional
    :param parser: The parser to use for BeautifulSoup, defaults to 'lxml'.
    :type parser: str, optional
    :param browse: Flag indicating whether to enter browse mode, defaults to False.
    :type browse: bool, optional
    :raises ValueError: If the URL is empty.
    :raises Exception: If there are issues with the request.
    :return: The content of the web page if successful, otherwise an error code.
    :rtype: Union[str, int]
    """
    return_string: str = ""
    page_index: list[str] = []
    curr_url: int = 0
    indextbl = "INDEX"

    if not header:
        options.add_argument("--headless")
    else:
        pass
    if proxy is not None:
        pass
    if log:
        pass

    with webdriver.Firefox(options=options) as driver:
        try:
            driver.get(url)
            WebDriverWait(driver, 4)
            body = WebDriverWait(driver, 10).until(
                EC.visibility_of_element_located((By.TAG_NAME, "body"))
            )
            soup = BeautifulSoup(body.get_attribute("outerHTML"), parser)
            anchor_tags = driver.find_elements(By.TAG_NAME, "a")

            urls = [
                tag.get_attribute("href")
                for tag in anchor_tags
                if tag.get_attribute("href") is not None
                and "javascript:" not in tag.get_attribute("href")
            ]
        except StaleElementReferenceException:
            c_log.warning(
                """Could not find element.
                [This is probably not asrch but the website] [See docs for potential fixed]
                Continuing"""
            )
    highlight_elements(soup)

    text_content = soup.body.get_text(separator="\n")
    try:
        for url_ in urls:
            curr_url += 1
            page_index.append(f"({curr_url}){url_}")
    except UnboundLocalError:
        c_log.error("Could not find page")

    if browse:
        formatted_string = ""
        text_content = f"""

        {text_content.strip().replace("\n\n", "").replace("", "").replace("None", "")}
            """

        Output.print(
            Box.create_box(
                f"Results {curr_url}",
                f"""{text_content})  
        """,
                f"\nINDEX\n{page_index}\nINDEX",
                padding=1,
                style="main",
            )
        )
        Alert.alert(1, "Finished")

        prompt = (
            "\n[q]uit, [h]istory ([<] back in history 1) " "or Enter page number: "
        )
        while page_num := input(prompt):
            if page_num not in ["q", "<"] and "h" not in page_num:
                page_num = page_num

                try:
                    print(f"Navigating to {urls[int(page_num) - 1]}")

                    history.append(urls[int(page_num) - 1])

                    if len(history) >= 2:
                        prev_history = (
                            history[-2]
                            .replace("https://", "")
                            .replace("www.", "")[:17]
                        )
                    else:
                        prev_history = ""
                    history_formatted = ", ".join(
                        [f"({i + 1}) {elem}" for i, elem in enumerate(history)]
                    )
                    formatted_string += (
                        f"{Colors.PASTEL_PURPLE}{prev_history}... "
                        f"{Colors.PASTEL_PINK}[<----] {Colors.PASTEL_MINT}[---->] "
                        f"{Colors.PASTEL_CYAN}{urls[int(page_num) - 1]}\n"
                        f"{Colors.PASTEL_BLUE}[HISTORY]\n"
                        f"{Colors.PASTEL_ORANGE}{history_formatted}\n"
                        f"{form.bar('.', '', t_color='', bg_color='')}"
                    )

                    print(formatted_string)

                    Alert.alert(2, "Loading...")

                    return urls[int(page_num) - 1]

                except ValueError:
                    c_log.error("Must be an int")
                    continue
                except IndexError:
                    c_log.error("Choose a correct number")
                    continue
                except UnboundLocalError:
                    c_log.error("Error finding page")
                    continue

            elif page_num == "q":
                return "q"
            elif "h" in page_num:
                history_val = page_num.replace("h", "")
                history_val = history_val
                try:
                    print(history[history_val])
                    return history[history_val]
                except IndexError:
                    c_log.warning(
                        "Please use proper value. Could not find history element"
                    )
                    continue
            elif page_num == "<":
                print(f"{colors.BOLD} Navigating back in history {colors.NC}")
                history_one = len(history) - 1
                return history[history_one]
            else:
                c_log.warning("Please enter proper value")
                continue

    else:
        return_string =f"""
        {indextbl}
        {text_content.strip().replace("\n\n\n\n", "").replace("\t", "").replace("None", "")}
        {indextbl}
        """
    return return_string
            
def get_image(
    url: str | None,
    proxy: Optional[str],
    header: bool,
    download: Optional[bool] = False,
    workspace: Optional[str] = None,
) -> Generator[str, None, None]:
    """Get the image from a URL.

    :param url: The URL of the image. If None, returns an empty generator.
    :param proxy: Proxy to be used for the request, defaults to None.
    :param header: Flag indicating whether to include headers in the request.
    :param download: Flag indicating whether to download the image, defaults to False.
    :raises ValueError: If the URL is empty.
    :raises NoSuchelementexception: If the element can't be found.
    :raises ElementNotVisibleException: If the element isn't visible in the DOM.
    :raises StaleElementReferenceException: If the element can no longer be accessed.
    :return: A generator yielding the image content as strings.
    """

    c_log.debug(url)
    c_log.debug(proxy)
    c_log.debug(header)

    if header:  # pragma: no cover
        c_log.info("Headless false")
    else:
        c_log.info("Headless true")
        options.add_argument("--headless")

    if proxy is not None:  # pragma: no cover
        c_log.info("Proxy set")

    with webdriver.Firefox(options=options) as driver:
        driver.get(url)
        WebDriverWait(driver, 4)

        try:
            image_is_visible = WebDriverWait(driver, 10).until(
                EC.visibility_of_element_located((By.TAG_NAME, "img"))
            )
            print(image_is_visible.text)
            c_log.debug("Found elements: <img>")

            if image_is_visible and download is False:
                images = driver.find_elements(By.TAG_NAME, "img")

                for image in images:
                    yield image.get_attribute("src")

            if image_is_visible and download is True:  # pragma: no cover
                images = driver.find_elements(By.TAG_NAME, "img")
                fname = f"{url}"
                head, _, _ = fname.partition(".com")
                filename = head.replace("https://www.", "").replace("/", " ")

                new_folder = f"images_{filename}_{formatted_date}{random.randint(1000, 5000)}"

                try:
                    os.mkdir(f"{resolved_path}/{workspace}/downloads/{new_folder}")
                except FileNotFoundError:
                    c_log.error("Error, trying to fix...")
                    os.mkdir(f"{resolved_path}/{workspace}/downloads/")
                    os.mkdir(f"{resolved_path}/{workspace}/downloads/{new_folder}")

                c_log.debug(f"Created new folder . . . {new_folder}")

                WebDriverWait(driver, 2)

                for _, image in enumerate(images):
                    src = image.get_attribute("src")
                    if src:
                        last_slash_index = src.rfind("/")
                        file_name = src[last_slash_index + 1 :]
                        time.sleep(3)
                        urllib.request.urlretrieve(
                            src, f"{resolved_path}/{workspace}/downloads/{new_folder}/{file_name[:12]}"
                        )
                        yield f"{Colors.GREEN}downloaded:{Colors.NC} {src} \
Downloaded to downloads/{resolved_path}/{src}"

        except NoSuchElementException:  # pragma: no cover
            c_log.error(
                "NoSuchElementException: Could not find search bar. Possible solutions:\n \
                + Change IP (you have been detected)\nCheck Internet \
                (element did not load in time)"
            )
            c_log.debug("|-> Continuing")

        except ElementNotVisibleException:  # pragma: no cover
            c_log.error("ElementNotVisibleException: Element not visible in DOM. ")
            c_log.debug("|-> Continuing")

        except StaleElementReferenceException:  # pragma: no cover
            c_log.error(
                Box.h2box(
                    None,
                    "Stale Element Reference Exception: Element cannot currently be accessed.",
                    None,
                ),
            )
            c_log.debug("|-> Continuing")

def get_screenshot(
    url: str, proxy: Optional[str], header: bool = False, log=True
) -> None:  # pragma: no cover
    """Take a screenshot of a web page.

    :param url: The URL of the web page.
    :type url: str
    :param proxy: Proxy to be used for the request, defaults to None.
    :type proxy: str, optional
    :param header: Flag indicating whether to include headers in the request, defaults to False
    :type header: bool
    :param log: Flag indicating whether to log actions, defaults to True.
    :type log: bool
    """

    c_log.debug(url)
    c_log.debug(proxy)
    c_log.debug(header)
    c_log.debug(log)

    if log:
        c_log.setLevel(logging.WARNING)
    else:
        c_log.setLevel(logging.DEBUG)

    if header:  # pragma: no cover
        c_log.info("Headless false")
    else:
        c_log.info("Headless true")
        options.add_argument("--headless")

    if proxy is not None:  # pragma: no cover
        c_log.info("Proxy set")

    with webdriver.Firefox(options=options) as driver:
        ipaddr = requests.get("https://checkip.amazonaws.com/", timeout=10).text
        c_log.debug("[✗]setting header")
        headers: dict[str, str] = {
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:84.0) \
            Gecko/20100101 Firefox/84.0",
        }

        c_log.debug(f"[✗]UA: {headers}")
        c_log.debug(f"[✗]IP: {ipaddr}")

        c_log.debug("Opening Driver")
        c_log.debug(f"Screnshotting {url}")

        WebDriverWait(driver, 2)
        driver.get(url)

        WebDriverWait(driver, 3)
        c_log.info("Taking Screenshot")
        c_log.debug("[D]Taking Screenshot")

        driver.save_screenshot("Screenshots/Screenshot-img.png")
        c_log.info("Screenshot saved to 'Screenshots' folder")

        c_log.info("Screenshot saved to 'Screenshots' folder")

def get_html(url: str, proxy: Optional[str], header: bool = False, log=True) -> str:
    """Get the HTML source code of a web page.

    :param url: The URL of the web page.
    :type url: str
    :param proxy: Proxy to be used for the request, defaults to None.
    :type proxy: str, optional
    :param header: Flag indicating whether to include headers in the request, defaults to False
    :type header: bool
    :param log: Flag indicating whether to log actions, defaults to True.
    :type log: bool
    :return: The HTML source code of the web page.
    :rtype: str
    """

    c_log.debug(url)
    c_log.debug(proxy)
    c_log.debug(header)
    c_log.debug(log)

    if log:
        c_log.setLevel(logging.WARNING)
    else:
        c_log.setLevel(logging.DEBUG)

    if header:  # pragma: no cover
        c_log.info("Headless false")
    else:
        c_log.info("Headless true")
        options.add_argument("--headless")

    with webdriver.Firefox(options=options) as driver:
        driver.get(url)
        time.sleep(3)
        WebDriverWait(driver, 3)
        return driver.page_source

def get_js(
    url: str, proxy: Optional[str], header: bool = False, log=True
) -> Generator[str, None, None]:
    """Get the JavaScript content from a web page.

    :param url: The URL of the web page.
    :type url: str
    :param proxy: Proxy to be used for the request, defaults to None.
    :type proxy: str, optional
    :param header: Flag indicating whether to include headers in the request, defaults to Fals.
    :type header: bool
    :param log: Flag indicating whether to log actions, defaults to True.
    :type log: bool
    :return: A generator yielding the JavaScript content as strings.
    :rtype: Generator[str, None, None]
    """

    c_log.debug(url)
    c_log.debug(proxy)
    c_log.debug(header)
    c_log.debug(log)

    if log:
        c_log.setLevel(logging.WARNING)
    else:
        c_log.setLevel(logging.DEBUG)

    pages: list[str] = []

    if header:  # pragma: no cover
        c_log.info("Headless false")
    else:  # pragma: no cover
        c_log.info("Headless true")
        options.add_argument("--headless")

    with webdriver.Firefox(options=options) as driver:
        driver.get(url)
        time.sleep(2)
        page = driver.find_elements(By.TAG_NAME, "script")

        for js in page:
            pages.append(js.get_attribute("src"))

        for jsurl in pages:
            if "google" not in jsurl and jsurl != "":
                driver.get(jsurl)
                time.sleep(2)
                body = WebDriverWait(driver, 10).until(
                    EC.visibility_of_element_located((By.TAG_NAME, "body"))
                )
                bar = form.bar("$", f"{jsurl}", Colors.RED, Colors.GREEN)
                print(bar)
                yield f"\n{body.text}"


def inspect(url: str, proxy: Optional[str], header: bool = False, log=True, parser = "html.parser") -> str:
    """Get the HTML source code of a web page.

    :param url: The URL of the web page.
    :type url: str
    :param proxy: Proxy to be used for the request, defaults to None.
    :type proxy: str, optional
    :param header: Flag indicating whether to include headers in the request, defaults to False
    :type header: bool
    :param log: Flag indicating whether to log actions, defaults to True.
    :type log: bool
    :return: The HTML source code of the web page.
    :rtype: str
    """

    if not header:
        options.add_argument("--headless")
    else:
        pass
    if proxy is not None:
        pass
    if log:
        pass

    with webdriver.Firefox(options=options) as driver:
        try:
            driver.get(url)
            WebDriverWait(driver, 4)
            body = WebDriverWait(driver, 10).until(
                EC.visibility_of_element_located((By.TAG_NAME, "body"))
            )
            soup = BeautifulSoup(body.get_attribute("outerHTML"), parser)
            anchor_tags = driver.find_elements(By.TAG_NAME, "a")

            urls = [
                tag.get_attribute("href")
                for tag in anchor_tags
                if tag.get_attribute("href") is not None
                and "javascript:" not in tag.get_attribute("href")
            ]
        except StaleElementReferenceException:
            c_log.warning(
                """Could not find element.
                [This is probably not asrch but the website] [See docs for potential fixed]
                Continuing"""
            )
    highlight_elements(soup)




"""
A module providing the 'open' command as an alternative to the Selenium based command.
This command utilizes the requests library for HTTP requests and BeautifulSoup for HTML parsing.
"""

from pathlib import Path
import os
import re
import logging
from typing import Optional, Literal
from urllib.parse import urljoin
import requests
from bs4 import BeautifulSoup, FeatureNotFound, Tag
import jsbeautifier

from asrch.modules.logging_formatter import ColorFormatter
from ascli.modules._alert import Alert
from ascli.modules._box import Box
from ascli.modules._output import Output
from ascli.modules._split import print_split_terminal
from ascli.utils.constants import Colors
from asrch.modules._formatting import Bar
from ascli.modules._highlighter import clean_and_format_urls as cln

history: list[str] = []  # session history for browse mode
tab_his:  list[str] = []
tabs_string: str = ""
tabs: list[str] = [] # tabs for browse mode sessions 
form = Bar()
colors = Colors()
c_log = logging.getLogger(__name__)
sh = logging.StreamHandler()

c_form = ColorFormatter("%(asctime)s|%(levelname)8s|%(message)s")

sh.setFormatter(c_form)

c_log.addHandler(sh)
config_path = Path(__file__).parent.parent.parent.parent.parent
path_with_config = config_path / '.config'

commands: list[str] = ["q", "<", "nt", "new_tab", "t", "sh", "show_history"]

def highlight_elements(soup: BeautifulSoup) -> BeautifulSoup:
    """
    Highlight all elements in BeautifulSoup object

    :param soup: The soup object to parse.
    :type: BeautifulSoup
    :return: soup
    :rtype: BeautifulSoup
    """

    for tag in soup.find_all(True):  
        if tag.string is not None and tag.name in ['div', 'hr']:
            tag.string = Box.create_box(None, tag.string, None, padding=1, style="h1u")

            separator_line = soup.new_tag('div')
            separator_line.string = "------------------------------------------------"
            try:
                tag.insert_before(separator_line)
                tag.insert_after(separator_line)
            except ValueError:
                continue
            
    for tag in soup.find_all("a"):
        if "href" in tag.attrs and tag["href"] is not None:
            tag["href"] = " * " + tag["href"]
        else:
            tag["href"] = " * "
    return soup

def get_index(url: str, mode: str = "") -> list[str]:
    """Fetches and processes URLs from a web page.

    :param url: The URL of the web page to fetch and process.
    :type url: str
    :param mode: Optional mode to determine the format of the output list. If `"url_list"`,
                 returns a list of URLs. Otherwise, returns a list with indexed URLs.
    :type mode: str, default is `""`
    :return: A list of URLs, either in indexed format or as a plain list depending on the mode.
    :rtype: list[str]
    :raises ValueError: If the provided URL is empty.
    :raises requests.exceptions.RequestException: If there is an error with the HTTP request.
    :raises FeatureNotFound: If BeautifulSoup cannot parse the HTML content.
    """
    headers = {"User-Agent": "Mozilla/5.0"}
    proxies = {} 

    if not url:
        raise ValueError("URL cannot be empty")

    try:
        response = requests.get(url, proxies=proxies, headers=headers)
        response.raise_for_status()
    except requests.exceptions.RequestException as e:

        print(f"Error occurred while fetching {url}: {str(e)}")
        return []

    try:
        soup = BeautifulSoup(response.content, "html.parser")
    except FeatureNotFound:
        return []

    href_list = []
    anchor_tags = soup.find_all("a")
    for anchor_tag in anchor_tags:
        if isinstance(anchor_tag, Tag):
            href = anchor_tag.get("href")
            if href:
                href_list.append(href)
                highlighted_content = anchor_tag.string
                if anchor_tag.string:
                    anchor_tag.string.replace_with(href)

    base_url = url
    url_list = []

    for url_ in href_list:
        modified_url = url_.replace("*", "").strip()  # Remove '*' and strip whitespace
        if modified_url.startswith("http"):
            url_list.append(modified_url)
        elif modified_url.startswith("/"):
            full_url = urljoin(base_url, modified_url)
            url_list.append(full_url)

    page_index = []
    try:
        for idx, url_ in enumerate(url_list, start=1):
            page_index.append(f"({idx}){url_}")
    except UnboundLocalError:
        print("Could not find page")
        
    if mode == "url_list": # this is absolutely fucking disgusting lets all ignore it :)
        return url_list
    else:
        return page_index

def handle_page_navigation_input(prompt: str, url_list: list[str], history: list[str], tabs: list[str], config_path: str) -> str:
    while True:
        page_num = input(prompt)
        try:
            if page_num not in commands and "h" not in page_num and "t" not in page_num:
                print(Colors.NC)
                try:
                    index = int(page_num) - 1
                    if 0 <= index < len(url_list):
                        print(f"Navigating to {url_list[index]}")
                        history.append(url_list[index])

                        prev_history = history[-2].replace("https://", "").replace("www.", "")[:17] if len(history) >= 2 else ""
                        history_formatted = ", ".join(f"({i + 1}) {elem}" for i, elem in enumerate(history))
                        formatted_string = f"{Colors.PASTEL_PURPLE}{prev_history}... {Colors.PASTEL_PINK}[<----] {Colors.PASTEL_MINT}[---->] {Colors.PASTEL_CYAN}{url_list[index]}\n{form.bar('.', '', t_color='', bg_color='')}"
                        
                        print(Colors.NC)
                        #print_split_terminal(left_text=f"* TABS\n{tabs_string}", right_text=f"* HISTORY\n{history}")
                        print(formatted_string)

                        Alert.alert(2, "Loading...")
                        return url_list[index]
                    else:
                        c_log.error(f"{index} Choose a correct number")
                except ValueError:
                    c_log.error("Must be an int")
                except IndexError as e:
                    print(e)
                    c_log.error("{page_num} Choose a correct number")
                except UnboundLocalError as e:
                    print(e)
                    c_log.error("Error finding page")
                except TypeError:
                    c_log.error("Bad input (maybe you put a url instead of int value?)")
            elif page_num == "q":
                print(Colors.NC)
                return "q"
            elif page_num.startswith("h"):
                try:
                    history_val = int(page_num[1:])
                    print(history[history_val])
                    return history[history_val]
                except IndexError:
                    c_log.warning("Please use proper value. Could not find history element")
            elif page_num in ["sh", "show_history"]:
                print(history)
            elif page_num == "<":
                try:
                    return history[-1]
                except IndexError:
                    c_log.warning("Please use proper value. Could not find history element")
            elif page_num in ["nt", "new_tab"]:
                get_url = input("url: ")
                tabs.append(get_url)
                tab_his.append(get_url)
                sub_index = []

                tabs_string = '\n'.join(f"{index + 1}. {tab}" for index, tab in enumerate(tabs))
                htab_string = '\n'.join(tab_his)
                tab_name = get_url.replace("https://", "")
                print(Colors.NC)
                print_split_terminal(left_text=f"* TABS\n{tabs_string}", right_text=f"* HISTORY\n{htab_string}")
                cfile_path = os.path.join(config_path, '.cache', f'{tab_name}.txt')
                if not os.path.exists(cfile_path):
                    try:
                        print(f"The file {cfile_path} does not exist in cache.")
                        with open(cfile_path, 'w') as f:
                            f.write(Box.create_box(
                                f"Results",
                                f"""{get_page(get_url)}""",
                                f"\n{get_index(get_url)}",
                                padding=1,
                                style="main",
                            ))
                        return get_url
                    except FileNotFoundError:
                        print(f"The file {cfile_path} does not exist in cache.")
                        return get_url
                else:
                   
                    search_string = "9b74c9897bac770ffc029102a200c5de7b5c6b4e715d0ab929c4a1c1f9f2b22d9eec09289c52ed0b2b38b4f3c7d5e2bd5d2a7cb63b223ce550501edbe2202"

                    with open(cfile_path, "r") as f:
                        found = False
                        for line in f:
                            if found:
                                # Clean the line and extract URLs
                                clean_line = line.replace(' ┋', '').replace('▢', '').replace('┉', '').strip()
                                urls = re.findall(r'https?://[^\s]+', clean_line)
                                sub_index.extend(urls)
                            elif search_string in line:
                                found = True

                    numbered_urls = [f"({idx}) {url}" for idx, url in enumerate(sub_index, start=1)]
                    Output.print(" ".join(numbered_urls))
                    page_num = 1

                    val = input(prompt)
                    if val == "nt":
                        get_url = input("Url: ")
                        result = get_url
                        return result
                    
                    try:
                        print(sub_index)

                        print(f"The URL at position {val} is: {result}")
                    except Exception as e:
                        print(e)
                        
                    print("return", result)
                    #return(result)   
            elif "t" in page_num and len(tabs) >= 1:
                try:
                    tab_value = int(page_num[1:])
                    print(f"{tabs[tab_value]}.txt")
                except IndexError:
                    c_log.error("Tab does not exist")
            else:
                c_log.warning("Please enter proper value")
        except TypeError:
            c_log.error(f"\"{page_num}\" Bad input (maybe you put a url instead of int value?)")


def get_page(
    url: str,
    header: str = "",
    proxy: Optional[dict[str, str]] = None,
    log: bool = True,
    *,
    parser: Optional[str] = "html.parser",
    browse: bool = False,
) -> str:
    """Get the content of a web page.

    :param url: The URL of the web page.
    :type url: str
    :param proxy: Proxy to be used for the request, defaults to None.
    :type proxy: Optional[str], optional
    :param log: Flag indicating whether to log the request, defaults to True.
    :type log: bool, optional
    :raises ValueError: If the URL is empty.
    :raises Exception: If there are issues with the request.
    :return: The content of the web page.
    :rtype: str
    """

    if not url:
        raise ValueError("URL cannot be empty")

    headers = {"User-Agent": "Mozilla/5.0"} if header else {}

    try:
        response = requests.get(url, proxies=proxy, headers=headers)
        response.raise_for_status()
    except requests.exceptions.RequestException as e:
        c_log.error(f"Error occurred while fetching {url}: {str(e)}")
        return ""

    try:
        soup = BeautifulSoup(response.content, parser)
        title_tags = soup.find_all("title")
        highlight_elements(soup)

        for title_tag in title_tags:
            if isinstance(title_tag, Tag):
                title_tag.string = title_tag.string

        page_index = get_index(url)
        url_list = get_index(url, "url_list")
        text_content = soup.text.strip().replace("\n\n\n", "").replace("", "")
        
        if not browse:
            return text_content
        else:
            formatted_string = ""
            text_content = f"""{text_content.replace("None", "")}"""
            Output.print(
                Box.create_box(
                    f"Results",
                    f"""{text_content}""",
                    f"\nINDEX\n{' '.join(page_index)}\nINDEX",
                    padding=1,
                    style="main",
                )
            )
            Alert.alert(1, "Finished")

            prompt = (
                f"\n{Colors.PASTEL_PINK}\n"
                f"[{Colors.MINT}{Colors.BOLD}q{Colors.NC}{Colors.PASTEL_PINK}]uit\n"
                f"[{Colors.MINT}{Colors.BOLD}h{Colors.NC}{Colors.PASTEL_PINK}]istory ([<] back in history 1)\n"
                f"[{Colors.MINT}{Colors.BOLD}nt{Colors.NC}{Colors.PASTEL_PINK}] create new tab\n"
                f"[{Colors.MINT}{Colors.BOLD}t{Colors.NC}{Colors.PASTEL_PINK}] go to nth tab\n"
                f"or Enter page number:"
            )
            
            return handle_page_navigation_input(prompt, url_list, history, tabs, config_path)
            
    except FeatureNotFound:
        c_log.error('Could not find LXML parser \n quickfix: "pip install lxml"')
        return ""

def get_html(
    url: str,
    header: str = "",
    proxy: Optional[dict[str, str]] = None,
    log: bool = True,
    *,
    parser: Optional[str] = "html.parser",
) -> BeautifulSoup:
    """Get the content of a web page.
    
    :param url: The URL of the web page.
    :type url: str
    :param proxy: Proxy to be used for the request, defaults to None.
    :type proxy: Optional[str], optional
    :param log: Flag indicating whether to log the request, defaults to True.
    :type log: bool, optional
    :raises ValueError: If the URL is empty.
    :raises Exception: If there are issues with the request.
    :return: The content of the web page.
    :rtype: str
    """
    if not url:
        raise ValueError("URL cannot be empty")

    headers = {"User-Agent": "Mozilla/5.0"} if header else {}

    try:
        response = requests.get(url, proxies=proxy, headers=headers)
        response.raise_for_status()
    except requests.exceptions.RequestException as e:
        raise Exception(f"Error occurred while fetching {url}: {str(e)}")
    try:
        soup = BeautifulSoup(response.content, parser)
    except FeatureNotFound:
        c_log.error('Could not find LXML parser \n quickfix: "pip install lxml"')
    return soup

def get_js(url: str, header: str = "", proxy: Optional[dict[str, str]] = None, log: bool = True, *, parser: Optional[str] = "html.parser") -> list[str] | str:
    """Get JavaScript sources from a web page."""
    jsoptions = jsbeautifier.default_options()
    if not url:
        raise ValueError("URL cannot be empty")

    headers = {"User-Agent": "Mozilla/5.0"} if header else {}

    try:
        response = requests.get(url, proxies=proxy, headers=headers)
        response.raise_for_status()
    except requests.exceptions.RequestException as e:
        raise Exception(f"Error occurred while fetching {url}: {str(e)}")

    try:
        soup = BeautifulSoup(response.content, parser)
    except FeatureNotFound:
        raise Exception('Could not find BeautifulSoup parser')

    script_tags = soup.find_all('script')

    js_sources: list[str]  = []
    for script in script_tags:
        src = script.get('src')
        if src:
            # Convert relative URLs to absolute URLs
            full_url = urljoin(url, src)
            
            js_sources.append(full_url)

    output_source: list[str] = []
    for js_source in js_sources:
        try:
            output_source.append(get_page(js_source))
        except ValueError as e:
            print("===========================================")
            c_log.error(e)
            print("===========================================")
        except KeyError as e:
            print("===========================================")
            c_log.error(e)
            print("===========================================")
            continue
    return output_source 

def inspect(
    url: str,
    header: str = "",
    proxy: Optional[dict[str, str]] = None,
    log: bool = True,
    *,
    parser: Optional[str] = "html.parser",
    browse: bool = False,
    mode: Literal["js", "html"] = "html",
) -> None:
    """
    Inspects a web page by making an HTTP request with optional custom headers and proxies.

    :param url: The URL of the web page to inspect.
    :type url: str
    :param header: Optional custom headers to include in the HTTP request.
    :type header: str, default is `''`
    :param proxy: Optional dictionary of proxy settings to use for the HTTP request. If `None`, no proxies are used.
    :type proxy: dict[str, str] | None, default is `None`
    :param log: Whether to enable logging of the HTTP request and response. Default is `True`.
    :type log: bool, default is `True`
    :param parser: Optional HTML parser to use with BeautifulSoup. Default is `'html.parser'`. If `None`, no parser is specified.
    :type parser: str | None, default is `'html.parser'`
    :param browse: Whether to enable browsing mode, which might affect how the page content is processed. Default is `False`.
    :type browse: bool, default is `False`
    :param mode: Specifies the mode of operation. Can be either `'js'` for JavaScript processing or `'html'` for HTML parsing.
    :type mode: Literal['js', 'html'], default is `'html'`
    
    :return: This function does not return a value. It performs actions based on the provided parameters.
    :rtype: None

    :raises ValueError: If the `url` is empty or invalid.
    :raises requests.exceptions.RequestException: If there is an issue with the HTTP request.
    :raises FeatureNotFound: If the specified parser is not found or cannot parse the content.
    :raises Exception: For other unexpected errors that may occur during the inspection process.

    :note: 
        - Ensure that the `header` parameter is properly formatted as a valid header string.
        - The `proxy` dictionary should be formatted with valid proxy settings.
        - The `mode` parameter determines whether JavaScript or HTML parsing is used.
        - The `browse` parameter might affect how the content is handled, depending on its implementation.

    :seealso:
        - `BeautifulSoup` documentation for parsing options: https://www.crummy.com/software/BeautifulSoup/bs4/doc/
        - `requests` documentation for HTTP request details: https://docs.python-requests.org/en/latest/
    """
    if not url:
        raise ValueError("URL cannot be empty")

    try:
        page_content = get_page(
            url=url, header=header, proxy=proxy, log=log, parser=parser
        )
        if mode == "html":
            inspect_content = get_html(
                url=url, header=header, proxy=proxy, log=log, parser=parser
            )

        if mode == "js":
            inspect_content = get_js(
                url=url, header=header, proxy=proxy, log=log, parser=parser
            )
        print_split_terminal(left_text=page_content, right_text=str(inspect_content))

    except ValueError as e:
        print(f"Error occurred during inspection: {str(e)}")

    return None



"""
Tests for selenium commands._send
"""

from asrch.commands._send import search_engines


def test_sends() -> None:
    """
    Tests get image function
    """

    assert "https" in search_engines(query="test", proxy=None, header=False)



"""
Send command
"""

import logging
import random
import time
from typing import Optional

from selenium import webdriver
from selenium.common.exceptions import (
    ElementNotVisibleException,
    NoSuchElementException,
    TimeoutException,
)
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.firefox.options import Options
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from asrch.modules.logging_formatter import ColorFormatter
from asrch.utils.constants import Colors

options = Options()


c_log = logging.getLogger(__name__)

sh = logging.StreamHandler()

c_form = ColorFormatter("%(asctime)s|%(levelname)8s|%(message)s")

sh.setFormatter(c_form)

c_log.addHandler(sh)

# CONSTANTS
DDG_URL: str = "https://lite.duckduckgo.com/html"
GGL_URL: str = "https://www.google.com/search"

url_index: list[str] = []



def scroll_down(driver) -> None:
    """Scrolls page until the end is reached

    :param driver: the driver to execute the script on
    :type driver: object
    :returns: None
    """
    c_log.debug("Scrolling")
    last_height = driver.execute_script("return document.body.scrollHeight")

    while True:
        driver.execute_script("window.scrollTo(0, document.body.scrollHeight);")

        WebDriverWait(driver, random.randrange(0, 2))

        new_height = driver.execute_script("return document.body.scrollHeight")

        if new_height == last_height:
            break

        last_height = new_height

def search_ddg(
    query: str, header: bool, proxy: Optional[str], log: bool
) -> list[str]:
    """Opens Google and searches for a query given by the user.

    :param url: The URL passed from the click argument, defaults to None
    :raises ValueError: If the URL is not provided
    :raises NoSuchelementexception: If the element can't be found.
    :raises ElementNotVisibleException: If the element isn't visible in the DOM.
    :return: The list of URLs from the webpage
    :rtype: list[str]
    """
    output: list[str] = [""]
    
    if log:
        c_log.setLevel(logging.DEBUG)
    else:
        c_log.setLevel(logging.WARNING)


    if header:  # pragma: no cover
        c_log.info("Headless false")
    else:
        c_log.info(
            "Headless true"
        )  # why does true mean no header LOL (not fixing it)
        options.add_argument("--headless")

    if proxy is not None:  # pragma: no cover
        c_log.info("Proxy set")

    with webdriver.Firefox(options=options) as driver:
        try:
            c_log.debug("Opening Driver")
            c_log.debug(DDG_URL)
            driver.get(DDG_URL)

            c_log.debug("Loooking for search bar")
            text_area = WebDriverWait(driver, 10).until(
                EC.visibility_of_element_located((By.CLASS_NAME, "search__input"))
            )

            WebDriverWait(driver, random.randrange(0, 2))
            text_area.send_keys(query)

            c_log.debug(f"Sending {query}")

            WebDriverWait(driver, random.randrange(0, 2))
            text_area.send_keys(Keys.ENTER)
            time.sleep(2) 

            lnks = driver.find_elements(By.XPATH, "//a[@href]")
            seen_urls = set()
            for lnk in lnks:
                _url = lnk.get_attribute("href")
                _title = lnk.get_attribute("text")
                
                
                if _url not in seen_urls:
                    seen_urls.add(_url)
                    url_index.append(_url)
                    output.append(
                        f"\033[1m {Colors.MINT}{_title.strip()}\033[0m\n{_url}\n"
                    )

            blacklist: list[str] = [
                "spreadprivacy",
                "javascript:",
                "bing",
                "bingj",
                "yahoo",
                "google",
                "duckduckgo",
            ]
            if not any(val in _url for val in blacklist):
                for urls in seen_urls:
                    seen_urls.add(urls)

        except NoSuchElementException as e:
            c_log.error(e)
            c_log.error("No element\n\t'-> continuing")

        except ElementNotVisibleException as e:
            c_log.error(e)
            c_log.error("Element not visible\n\t'-> continuing")

        except TimeoutException as e:
            c_log.error(e)
            c_log.error("Timed out\n\t'-> continuing")

        return output

def search_ggl(query: str, header: bool, proxy: Optional[str], log: bool) -> str:
    """Opens Google and searches for a query given by the user.

    :param url: The URL passed from the click argument, defaults to None
    :raises ValueError: If the URL is not provided
    :raises NoSuchelementexception: If the element can't be found.
    :raises ElementNotVisibleException: If the element isn't visible in the DOM.
    :raises Timeoutexception: If seleniumm times out waiting for element.
    :return: The list of URLs from the webpage
    :rtype: list[str]
    """
    output: list[str] = [""]

    if log:
        c_log.setLevel(logging.DEBUG)
    else:
        c_log.setLevel(logging.WARNING)

    if header:  # pragma: no cover
        c_log.info("Headless false")
    else:
        c_log.info(
            "Headless true"
        )  # why does true mean no header LOL (not fixing it)
        options.add_argument("--headless")

    if proxy is not None:  # pragma: no cover
        c_log.info("Proxy set")

    with webdriver.Firefox(options=options) as driver:
        try:
            c_log.debug("Opening Driver")
            c_log.debug(GGL_URL)
            driver.get(GGL_URL)
            c_log.debug("Loooking for search bar")
            text_area = WebDriverWait(driver, 10).until(
                EC.visibility_of_element_located((By.CLASS_NAME, "gLFyf"))
            )
            WebDriverWait(driver, 2)
            text_area.send_keys(query)

            c_log.debug(f"Sending {query}")

            WebDriverWait(driver, random.randrange(0, 2))
            text_area.send_keys(Keys.ENTER)
            time.sleep(3)  
            lnks = driver.find_elements(By.XPATH, "//a[@href]")
            curr_url = 0
            seen_urls = set()
            i = 0
            for lnk in lnks:
                _url = lnk.get_attribute("href")
                _title = lnk.get_attribute("text")  # Get the title attribute
                i += 1

                if _url not in seen_urls:
                    seen_urls.add(_url)
                    output.append(
                        f"\033[1m {Colors.MINT}{_title.strip()}\033[0m\n{_url}\n"
                    )

        except NoSuchElementException as e:
            c_log.error(e)
            c_log.error("No element\n\t'-> continuing")

        except ElementNotVisibleException as e:
            c_log.error(e)
            c_log.error("Element not visible\n\t'-> continuing")

        except TimeoutException as e:
            c_log.error(e)
            c_log.error("Timed out\n\t'-> continuing")

        return output


def search_engines(
        query: str, header: bool = False, proxy: Optional[str] = None, log: bool = True, browse = False
) -> str | int:
    output: list[str] = []

    ddg_links = search_ddg(query, header=header, proxy=proxy, log=log)
    output.extend(ddg_links)
    
    
    # Collect links from Google
    ggl_links = search_ggl(query, header=header, proxy=proxy, log=log)
    output.extend(ggl_links)

    if browse:
        for index, line in enumerate(output, start=1):
            print(f"{index}. {line}")
            
        prompt = (
            "\n[q]uit, [h]istory ([<] back in history 1) " "or Enter page number: "
        )
        
        while page_num := int(input(prompt)):
            if page_num not in ["q", "<"]:
                print(f"Navigating to {url_index[page_num - 2]}")
                formatted_string = (
                        f"{Colors.PASTEL_PINK}[<----] {Colors.PASTEL_MINT}[---->] "
                        f"{Colors.PASTEL_CYAN}{url_index[int(page_num) - 1]}\n"
                        f"{Colors.PASTEL_BLUE}[HISTORY]\n"
                    )

                print(formatted_string)
                return (url_index[page_num - 2])


    else:
        return "\n".join(output)
        



"""
Find and return specified element with given URL
"""

import logging
from typing import Optional

from selenium import webdriver
from selenium.common.exceptions import NoSuchElementException, TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.firefox.options import Options
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from asrch.modules._formatting import Bar
from asrch.modules.logging_formatter import ColorFormatter

options = Options()


c_log = logging.getLogger(__name__)

sh = logging.StreamHandler()

c_form = ColorFormatter("%(asctime)s|%(levelname)8s|%(message)s")

sh.setFormatter(c_form)

c_log.addHandler(sh)



def find_elements(
    url: str,
    locator: str,
    element: str,
    header: bool = False,
    proxy: Optional[str] = None,
    log=True,
) -> str:
    """
    Find elements based on given locator + element name
    The possible locators are: TAG_NAME, CLASS_NAME, ID, XPATH


    :param url: The URL of the web page.
    :type url: str
    :param element: The element to return
    :type element: str
    :param proxy: Proxy to be used for the request, defaults to None.
    :type proxy: Optional[str], optional
    :param header: Flag indicating whether to include headers in the request, defaults to False.
    :type header: bool, optional
    :raises ValueError: If the URL is empty.
    :raises Exception: If there are issues with the request.
    :return: The content of the web page.
    :rtype: str
    """

    c_log.debug(url)
    c_log.debug(element)
    c_log.debug(proxy)
    c_log.debug(header)
    c_log.debug(log)

    if log:
        c_log.setLevel(logging.WARNING)
    else:
        c_log.setLevel(logging.DEBUG)

    if header:  # pragma: no cover
        c_log.info("Headless false")
    else:
        c_log.info("Headless true")
        options.add_argument("--headless")

    if proxy is not None:  # pragma: no cover
        c_log.info("Proxy set")

    with webdriver.Firefox(options=options) as driver:
        output: str = ""
        c_log.info(f"Looking for element {element}")
        driver.get(url)
        try:
            if locator.lower() in ["tag", "tag_name"]:  # TAG NAME
                body = WebDriverWait(driver, 10).until(
                    EC.visibility_of_all_elements_located((By.TAG_NAME, element))
                )
            if locator.lower() in ["class", "class_name"]:  # CLASS NAME
                body = WebDriverWait(driver, 10).until(
                    EC.visibility_of_all_elements_located((By.CLASS_NAME, element))
                )
            if locator.lower() in ["id"]:  # ID
                body = WebDriverWait(driver, 10).until(
                    EC.visibility_of_all_elements_located((By.ID, element))
                )
            if locator.lower() in ["xpath", "x"]:  # XPATH
                body = WebDriverWait(driver, 10).until(
                    EC.visibility_of_all_elements_located((By.XPATH, element))
                )
            else:
                c_log.error("No valid locator used...")

        except NoSuchElementException:
            c_log.error(f"Could not find {element}")
        except TimeoutException:
            c_log.error(f"Timed out. Could not find {element} in 10s")
        try:
            for elem in body:
                output += Bar.bar("-", "Element", "", "") + "\n"
                output += str(elem) + "\n"
                output += Bar.bar("-", "Element", "", "") + "\n"
                output += elem.text + "\n"
        except UnboundLocalError:
            c_log.error(f"Please try again {locator.lower()}")
        return output



"""
Entry point for asrch

Main file for asrch which handles all modules
CLI argument and command handling
"""

import logging
import sys
from typing import Optional
from datetime import datetime
import os

import click
import typer
import typer.core
from typing_extensions import Annotated
import toml
from pathlib import Path

import asrch.webdriver.webdriver_manager as webdriver_manager
from asrch.commands._find import find_elements
from asrch.commands._open import get_page, get_html, get_image, get_js, get_screenshot
from asrch.commands._send import search_engines
from asrch.commands.no_driver._nd_open import get_page, inspect, get_js
from asrch.modules._formatting import Bar as Sep
from asrch.modules.logging_formatter import ColorFormatter
from asrch.utils.constants import Colors, OpenModes  # , ConfigDefaults

from ascli.modules._output import Output


form = Sep()
colors = Colors()

app = typer.Typer()

config_path = Path(__file__).parent
config = toml.load(Path(f"{config_path}/config/undetected.toml"))
ws_config = toml.load(Path(f"{config_path}/config/current_ws.toml"))
workspace_path = Path(f"{config_path}/config/config/workspaces.txt")
cworkspace_path = Path(f"{config_path}/config//current_ws.txt")

config_path = Path(__file__).parent.parent.parent
path_with_config = config_path / '.cache'


current_dir = os.path.dirname(os.path.realpath(__file__))

resolved_path = f"{Path(__file__).parent.parent.parent}/workspaces/"

current_datetime = datetime.now()

c_log = logging.getLogger(__name__)
c_log.setLevel(logging.DEBUG)


sh = logging.StreamHandler()

c_form = ColorFormatter("%(asctime)s|%(levelname)8s|%(message)s")


sh.setFormatter(c_form)

c_log.addHandler(sh)

def ensure_ws_doesnt_exist(ws_name: str) -> bool:
    file_path = Path("src/asrch/config/workspaces.txt")
    with open(file_path, 'r') as f:
        for line in f:
            if f"name:{ws_name}" in line.strip():
                return False  
    return True 

def create_workspace(name: str, current_datetime: str, config: str):
    if ensure_ws_doesnt_exist(name):
        with open(Path("src/asrch/config/workspaces.txt"), 'a') as f:
            f.write(f"""\n\n
            --------------
            WORKSPACE {name}
            --------------
            name:{name}
            config: {config} 
            created:{current_datetime}
            --------------
            END WORKSPACE {name}
            --------------""".replace("                ", ""))
            os.mkdir(f"{resolved_path}/{name}/")
        typer.echo(f"Workspace '{name}' created successfully.")
    else:
        c_log.error(f"Workspace '{name}' already exists.")
        
def get_ws():
    with open("src/asrch/config/current_ws.txt", 'r') as f:
        for line in f:
            return

@app.command(name="open")
def open_(
    mode: OpenModes = typer.Argument(help="Mode for command"),
    url: str = typer.Argument(help="URL to retrieve"),
    proxy: str = typer.Option(
        None, "--proxy", "-p", help="Proxy to send request <ip:port> [optional]"
    ),
    header: bool = typer.Option(False, "--header", "-H", help="Show browser header"),
    browse: bool = typer.Option(
        False,
        "--browse",
        "-b",
        help="Enable browsing (using keyboard inputs to open urls)",
    ),
    pager: bool = typer.Option(False, "--pager", "-P", help="Output [JS] in pager"),
    download: bool = typer.Option(
        False, "--download", "-d", help="Download all scraped images"
    ),
    log: bool = typer.Option(
        False, "--log", "-l", help="Suppress all logs (for emacs mode)"
    ),
    nodriver: bool = typer.Option(
        False,
        "--nodriver",
        "-n",
        help="Use requests and BS4 instead of selenium (faster but more detectable)",
    ),
    parser: str = typer.Option(
        "html.parser",
        "--parser",
        help="HTML parser to use for nodriver. Choices: html.parser, lxml",
    ),
    silent: bool = typer.Option(
        False,
        "--silent",
        "-s",
        help="Output to text file instead of terminal (defaults to workspace folder)",
    ),
    inspect_mode: str = typer.Option(
        "html",
        "-iM",
        help = "What part of the page you would like to inspect. :Note: -M js can have a LONG ASS output (~5k line history rec)",
    ),
        
):
    """Open a URL and perform various operations.

    :param mode: Mode for the command.
    :type mode: OpenModes

    :param url: URL to retrieve.
    :type url: str

    :param proxy: Proxy to send request <ip:port> [optional].
    :type proxy: str, optional

    :param header: Show browser header.
    :type header: bool, optional

    :param browse: Enable browsing (using keyboard inputs to open URLs).
    :type browse: bool, optional

    :param pager: Output [JS] in pager.
    :type pager: bool, optional

    :param download: Download all scraped images.
    :type download: bool, optional

    :param log: Suppress all logs (for emacs mode).
    :type log: bool, optional

    :param nodriver: Use requests and BS4 instead of selenium (faster but more detectable).
    :type nodriver: bool, optional

    :param parser: HTML parser to use for nodriver. Choices: html.parser, lxml.
    :type parser: str, optional

    :param silent: Output to text file instead of terminal (defaults to workspace folder).
    :type silent: bool, optional

    :param inspect_mode: What part of the page you would like to inspect.
    :type inspect_mode: str, optional

    :Note:
        - For `log`, this option is intended for emacs mode and should be ignored in normal CLI mode, but can be used if needed.
        - `inspect_mode` with `-M js` can produce a large output (~5k line history record).
    """
    if not nodriver:
        with webdriver_manager.format_alerts(
            2,
            f"[{get_ws()}] Begin {mode.value} | {'-b' if browse else ''} | {'-h' if header else ''} | {url}",
        ):
            if mode.value == "text":
                if browse:
                    returned_url = url
                    while returned_url not in ["q", "<"]:
                        returned_url = get_page(
                            url=returned_url,
                            proxy=proxy,
                            header=header,
                            log=log,
                            browse=browse,
                        )

                elif browse is False:
                    if silent: 
                        with open(f"{resolved_path}/{get_ws()}/output.txt", "w") as f:
                            f.write(f"{datetime} \n")
                            f.write((get_page(url=url, proxy=proxy, header=header, log=log)))
                    else:
                        print(get_page(url=url, proxy=proxy, header=header, log=log))

            elif mode.value == "body":
                print(get_page(url=url, proxy=proxy, header=header))
            elif mode.value == "html":
                print(get_html(url=url, proxy=proxy, header=header))
            elif mode.value == "js":
                js_srcs = get_js(url=url, proxy=proxy, header=header)
                if log:
                    for js_src in js_srcs:
                        print(js_src)
                elif pager:
                    click.echo_via_pager(js_srcs)
                else:
                    for js_src in js_srcs:
                        print(js_src)
            elif mode.value == "img":
                for images in get_image(url, proxy, header, download=download, workspace=get_ws()):
                    print(images)
            elif mode.value == "ss":
                get_screenshot(url, proxy, header)
                click.echo("SCREENSHOTTED")
    else:  # nodriver
        with webdriver_manager.format_alerts(
            2,
            f"Begin {mode.value} | {'-b' if browse else ''} | {'-h' if header else ''} | {url}",
        ):           
            if mode.value == "html":
                print(get_html(url, parser=parser))
            elif mode.value in ["ins", "inspect"]:
                if pager:
                    click.echo_via_pager(inspect(url=url,proxy=proxy, header=header,mode=inspect_mode))
                else:
                    inspect(url=url,proxy=proxy, header=header,mode=inspect_mode)
            elif mode.value == "text":
                if browse:
                    returned_url = url
                    while returned_url not in ["q", "<"]:
                                returned_url = get_page(
                                    url=returned_url,
                                    proxy=proxy,
                                    header=header,
                                    log=log,
                                    browse=browse,
                                )

            elif mode.value == "js":
                print(get_js(
                    url=url,
                    proxy=proxy,
                    header=header,
                    log=log,
                    ))
@app.command()
def browse(
    url = "https://asrch.bitbucket.io",
    proxy: str = typer.Option(
        None, "--proxy", "-p", help="Proxy to send request <ip:port> [optional]"
    ),

    log: bool = typer.Option(
        False, "--log", "-l", help="Suppress all logs (for emacs mode)"
    ),
    nodriver: bool = typer.Option(
        False,
        "--nodriver",
        "-n",
        help="Use requests and BS4 instead of selenium (faster but more detectable)",
    ),
    parser: str = typer.Option(
        "html.parser",
        "--parser",
        help="HTML parser to use for nodriver. Choices: html.parser, lxml",
    ),        
):
    """Browse the web.

    :param url: The URL to start browsing from (default is "https://asrch.bitbucket.io").
    :type url: str

    :param proxy: Proxy to send requests (<ip:port>) [optional].
    :type proxy: str, optional

    :param log: Suppress all logs (for emacs mode).
    :type log: bool, optional

    :param nodriver: Use requests and BS4 instead of selenium (faster but more detectable).
    :type nodriver: bool, optional

    :param parser: HTML parser to use for nodriver. Choices: html.parser, lxml (default is html.parser).
    :type parser: str, optional

    :return: None
    :rtype: None

    :raises ValueError: If invalid parser option is provided.

    :Example:

        To browse using default settings:

        >>> browse()

        To browse using a proxy and suppress logs:

        >>> browse(proxy="127.0.0.1:8080", log=True)

        To browse without using a webdriver and specify a parser:

        >>> browse(nodriver=True, parser="lxml")
    """
    browse = True
    if browse:
        returned_url = url
        
        while returned_url not in ["q", "<"] and returned_url:
            returned_url = get_page(
                   url=returned_url,
                   proxy=proxy,
                   log=log,
                   browse=browse,
               )


@app.command()
def ccache():
    """
    Clear browser cache
    
    """
    for file in path_with_config.iterdir():
        if file.is_file() and file.suffix == '.txt':
            c_log.debug(f"del: {file}")
            file.unlink()

            print(f"{Colors.GREEN}Cache cleared")
@app.command()
def search(
    query: str = typer.Argument(help="Search query"),
    proxy: str = typer.Option(
        "", "--proxy", "-p", help="Proxy to send request <IP:port> [optional]"
    ),
    browse: bool = typer.Option(False, "--browse", "-b", help="Enable browse mode"),
    header: bool = typer.Option(False, "--header", "-H", help="Show browser header"),
    log: bool = typer.Option(
        False, "--log", "-l", help="Suppress all logs (for emacs mode, ignored in CLI)"
    ),
):
    """Search function to perform a search operation.

    :param header: Show browser header. Annotated with bool.
    :type header: Annotated[bool, typer.Option(help="show browser header")]

    :param proxy: Proxy to send request <IP:port> [optional]. Annotated with str.
    :type proxy: Annotated[str, typer.Option(help="proxy to send request <IP:port> [optional]")]

    :param log: toggle logging message visibility
    :type log: bool
    :default: false
    : This flag is intended for the emacs plugin and is not made to
      be used within the normal CLI mode however you can use it if
      you like.
    """

    with webdriver_manager.format_alerts(
        2,
        f"Begin | {query}",
    ):        

        returned_url: str | int = search_engines(query, header, proxy, log, browse)
        if browse:
            while returned_url:
                returned_url = get_page(
                    url = returned_url,
                    proxy=proxy,
                    header=header,
                    log=log,
                    browse=browse,
                )
        else:
            print(search_engines(query, header, proxy, log, browse))

            

@app.command()
def find(
    url: Annotated[str, typer.Option(help="URL to retrieve")],
    element: Annotated[str, typer.Option(help="Element to return")],
    proxy: Annotated[
        Optional[str], typer.Option(help="proxy to send request <ip:port> [optional]")
    ] = "",
    header: Annotated[bool, typer.Option(help="show browser header")] = False,
    log: Annotated[
        bool,
        typer.Option(
            help="supress all logs (this option is for emacs mode and should be ignore in the CLI)"
        ),
    ] = False,
    locator: Annotated[str, typer.Option(help="Locator to find element")] = "tag_name",
):
    """Find an element on a web page.

    :param url: URL to retrieve.
    :type url: str

    :param element: Element to return.
    :type element: str

    :param proxy: Proxy to send request (<ip:port>) [optional].
    :type proxy: Optional[str]
    :default proxy: ""

    :param header: Show browser header.
    :type header: bool
    :default header: False

    :param log: Suppress all logs.
    :type log: bool
    :default log: False

    :param locator: Locator to find the element.
    :type locator: str
    :default locator: "tag_name"

    :return: None
    :rtype: None
    """
    with webdriver_manager.format_alerts(
        2,
        f"Begin | {url}"):
        print((find_elements(url, locator, element, header, proxy, log)))


@app.command()
def workspace(
    initialize: bool = typer.Option(False, "--initialize", help="Initialize workspace"),
    create: bool = typer.Option(False, "--create", help="Create workspace"),
    config: str = typer.Option(None, "-c" ,"--conf", help="The name of the workspace to act on"),
    name: str = typer.Option(None,"-n", "--name", help="The name of the workspace to act on"),
    delete: bool = typer.Option(False, "--delete", help="Delete workspace")
):
    """Perform operations related to workspaces.
    Depending on the options provided, this function can initialize, create, delete,
    or perform other actions related to workspaces.

    :param initialize: Flag to initialize workspace.
    :type initialize: bool
    
    :param create: Flag to create workspace.
    :type create: bool
    
    :param config: The name of the workspace to act on.
    :type config: str
    
    :param name: The name of the workspace to act on.
    :type name: str
    
    :param delete: Flag to delete workspace.
    :type delete: bool
    """
    if initialize:
        if not ensure_ws_doesnt_exist(name):
            with open("src/asrch/config/current_ws.txt", 'w') as f:
                f.write(name)
            typer.echo(f"{Colors.MINT}Initialized workspace: {name} and {config}")
            sys.ps1 = name

        else:
            typer.echo("Error: Please provide an existing workspace name with --name option.")

    if create:
        if name:
            current_datetime = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
            create_workspace(name, current_datetime, config)
        else:
            typer.echo("Error: Please provide a workspace name with --name option.")

@app.command()
def conf():
    """Output current config
    """
    print(f"{Colors.MINT}{Colors.BOLD}\n[Workspace]{Colors.NC}")
    with open("src/asrch/config/current_ws.txt", 'r') as f:
        for line in f:
            print(line)
        
    print(f"{Colors.PASTEL_PINK}{Colors.BOLD}[basic]{Colors.NC}")
    print(f"url = {config['basic']['url']}\n")

    print(f"{Colors.BOLD}{Colors.PASTEL_PINK}[privacy]{Colors.NC}")
    print(f"proxy = {config['privacy']['proxy']}\n")

    print(f"{Colors.PASTEL_PINK}{Colors.BOLD}[commands]{Colors.NC}\n")

    commands = config['commands']
    for command, settings in commands.items():
        print(f"  {Colors.PASTEL_ORANGE}[commands.{command}]{Colors.NC}")
        for setting, value in settings.items():
            print(f"{setting} = {value}")
        print() 


if __name__ == "__main__":
    app()




"""
Tests for find command
"""

import pytest
from selenium import webdriver

from asrch.commands._find import find_elements




def test_id() -> None:
    """
    Tests id function
    """

    assert "Countries of the World" in (
        find_elements(
            "https://www.scrapethissite.com/pages/simple",
            "ID",
            "page",
            header=False,
            proxy=None,
        )
    )


def test_tag_name() -> None:
    """
    Tests tag_name function
    """

    assert "Scrape This Site" in (
        find_elements(
            "https://www.scrapethissite.com/pages/simple",
            "TAG_NAME",
            "a",
            header=False,
            proxy=None,
        )
    )


def test_class_name() -> None:
    """
    Tests class_name function
    """

    assert "Countries of the World" in (
        find_elements(
            "https://www.scrapethissite.com/pages/simple",
            "CLASS_NAME",
            "row",
            header=False,
            proxy=None,
        )
    )


def test_xpath() -> None:
    """
    Tests xpath function
    """

    assert "Countries of the World" in (
        find_elements(
            "https://www.scrapethissite.com/pages/simple",
            "XPATH",
            "/html/body/div/section/div/div[1]/div/h1",
            header=False,
            proxy=None,
        )
    )



"""
Formatting module for asrch
"""

import re
import shutil
from typing import Optional

from asrch.utils.constants import Colors

colors = Colors()
terminal_size = round(shutil.get_terminal_size().columns)


class Bar:
    """
    Bar class to create bar object
    """

    @staticmethod
    def bar(
        char: str,
        title: str,
        t_color: Optional[Colors | str],
        bg_color: Optional[Colors | str],
    ) -> str:
        """
        :param char: The character used to represent the bar
        :type char: str

        :param title: The title of the bar
        :type title: str

        :param title_color: The color of the title. Can be a Colors type or a string representing
        a color.
        :type title_color: BarColor | str

        :param background_color: The background color of the bar. Can be a BarColor enum or a
        string representing a color.
        :type background_color: Color | str
        """

        terminal_size = round(shutil.get_terminal_size().columns)
        bar_array: list[str] = []

        title = f"{t_color}{title:^{round(terminal_size)}}{bg_color}"
        if len(char) < 1:
            raise ValueError("Char must be 1 character long")

        for _ in range(terminal_size):
            bar_array.append(char)
        formatted_array = (
            str(bar_array)
            .replace("'", "")
            .replace(",", "")
            .replace(" ", "")
            .replace("[", "")
            .replace("]", "")
        )
        if bg_color is None and t_color is None:
            return f"{formatted_array}\
\n{title:{char}^{round(terminal_size)}}\
\n{formatted_array}".replace("None", "")
        else:
            return f"{bg_color}{formatted_array}\
\n{title:{char}^{round(terminal_size)}}{Colors.NC}\
\n{bg_color}{formatted_array}{Colors.NC}"


class Highlighter:
    """
    Highlighter class for methods regarding string formatting
    """

    @staticmethod
    def highlight_html(text: str) -> str:
        """
        Highlight certain keywords to make output more readable

        :param text: The string to highlight
        :type text str
        """

        text = re.sub(r"var", rf"{Colors.ORANGE}", text)
        text = re.sub(r"function", rf"{Colors.ORANGE}", text)
        text = re.sub(r"\d+", rf"{Colors.GREEN}", text)
        return text

    @staticmethod
    def highlight_page_output(text: str) -> str:
        """
        Highlight certain keywords to make output more readable

        :param text: The string to highlight
        :type text: str
        :return: The string with highlighted keywords
        :rtype: str
        """

        # Define color escape sequences or color codes
        colors = {
            "URL": Colors.DIM_GRAY,
            "HEADER": Colors.PASTEL_BLUE,
            "CODE": Colors.PASTEL_GREEN,
            "COMMAND": "\x1b[48;5;238m",
            "NUMBER": Colors.PASTEL_CYAN,
        }

        # Replace URLs
        """text = re.sub(
            r"https?://\S+",
            lambda match: f"{colors['URL']}{match.group(0)}\x1b[0m",
            text,
        )"""

        # Highlight headers
        text = re.sub(
            r"(#+)\s*(.*?)\s*(#+)*",
            lambda match: f"{colors['HEADER']}{match.group(1)}{match.group(2)}{match.group(3) if match.group(3) else ''}\x1b[0m",
            text,
        )

        # Highlight code snippets (assuming they are in backticks)
        text = re.sub(
            r"`(.*?)`",
            lambda match: f"{colors['CODE']}`{match.group(1)}`\x1b[0m",
            text,
        )

        text = re.sub(
            r"^(>>> .*)$",
            lambda match: f"{colors['COMMAND']}{match.group(1)}\x1b[0m",
            text,
            flags=re.MULTILINE,
        )

        return text

    @staticmethod
    def highlight_nums(text: str) -> str:
        # Define ANSI escape codes for bold pink text
        ANSI_BOLD_PINK = "\033[1;95m"
        ANSI_RESET = "\033[0m"

        # Regex pattern to find numbers (including decimals and negative signs)
        pattern = r"[-+]?\d*\.\d+|\d+"

        # Replace all matches with bold pink text
        highlighted_text = re.sub(pattern, f"{ANSI_BOLD_PINK}\\g<0>{ANSI_RESET}", text)

        return highlighted_text


class Text:
    def print_centered(text, terminal_width=80):
        # Split text into lines
        lines = text.splitlines()

        # If there are no lines, return an empty string
        if not lines:
            return ""

        # Calculate left indent to center the text
        left_indent = (terminal_width - max(len(line) for line in lines)) // 2

        # Center each line within terminal width
        centered_lines = [line.center(terminal_width) for line in lines]

        # Join centered lines with newline characters and return as a single string
        centered_text = "\n".join(centered_lines)

        return centered_text

    def print_indented(text, indentation="\t\t"):
        lines = text.splitlines()
        new_output = []
        i = 0
        for line in lines:
            i += 1
            formatted_line = line
            formatted_line = f"{i}│{line}"
            new_output.append(formatted_line)

        indented_text = "\n".join(new_output)
        return indented_text



import time
import sys
import shutil
from contextlib import contextmanager
from ascli.modules._alert import Alert
from ascli.modules._output import clear_last_lines, move_cursor_to_bottom, save_cursor_position, restore_cursor_position
from asrch.modules._formatting import Bar
from asrch.utils.constants import Colors

colors = Colors()
form = Bar()


@contextmanager
def format_bars(bar_char, message, color_start, color_end):
    start_time = time.time()
    print(form.bar(bar_char, "START " + message, color_start, colors.TIME))
    yield
    end_time = time.time()
    elapsed_time = end_time - start_time
    print(
        form.bar(
            bar_char,
            f"END {message} (Time: {elapsed_time:.2f} seconds)",
            color_end,
            colors.TIME,
        )
    )


@contextmanager
def format_alerts(delay, msg):
    print("\033[H\033[2J")
    start_time = time.time()
    Alert.alert(delay=delay,msg=msg)
    print()
    yield
    print()
    end_time = time.time()
    elapsed_time = end_time - start_time
    print("\033[H\033[2J")
    Alert.alert(delay=delay+3,msg=f"Done ({elapsed_time})")
    print("\033[H\033[2J")




import logging
from inspect import stack
from pathlib import Path

from asrch.modules._formatting import Highlighter
from asrch.utils.constants import Colors, Files

current_file = Files()

highlight_text = Highlighter()
color = Colors()
nc = color.NC
h = ""


class ColorFormatter(logging.Formatter):
    def __init__(self, fmt):
        super().__init__()
        fmt = fmt.split("|")
        self.FORMATS = {
            logging.DEBUG: f"{color.DIM_GRAY}{color.GRAY}{Path(stack()[1].filename).name:<22}↪{color.TIME}{fmt[0]}{color.NC}{color.YELLOW}{fmt[1]}{color.DIM_GRAY}{h:<4}%(lineno)d↪ {nc}{fmt[2]} {nc}",
            logging.INFO: f"{color.DIM_GRAY}{color.GRAY}{Path(stack()[1].filename).name:<22}↪{color.TIME}{fmt[0]}{color.NC}{color.BLUE}{fmt[1]}{color.DIM_GRAY}{h:<4}%(lineno)d↪ {nc}{fmt[2]} {color.NC}",
            logging.WARNING: f"{color.DIM_GRAY}{color.GRAY}{Path(stack()[1].filename).name:<22}↪{color.TIME}{fmt[0]}{color.NC}{color.ORANGE}{fmt[1]}{color.DIM_GRAY}{h:<4}%(lineno)d↪ {nc}{fmt[2]} {color.NC}",
            logging.ERROR: f"{color.DIM_GRAY}{color.GRAY}{Path(stack()[1].filename).name:<22}↪{color.TIME}{fmt[0]}{color.NC}{color.RED}{fmt[1]}{color.DIM_GRAY}{h:<4}%(lineno)d↪ {nc}{fmt[2]} {color.NC}",
            logging.CRITICAL: f"{color.DIM_GRAY}{color.GRAY}{Path(stack()[1].filename).name:<22}↪{color.TIME}{fmt[0]}{color.NC}{color.CRITICAL}{fmt[1]}{color.DIM_GRAY}{h:<4}%(lineno)d↪ {nc}{fmt[2]} {color.NC}",
        }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)



"""
Tests for no driver open command
"""

from unittest.mock import patch

import pytest
from bs4 import FeatureNotFound

from asrch.commands.no_driver._nd_open import ND_Open


@pytest.fixture
def mock_response():
    class MockResponse:
        def __init__(self, content, status_code):
            self.content = content
            self.status_code = status_code

        def raise_for_status(self):
            pass

    return MockResponse(b"<html><body><p>Hello World</p></body></html>", 200)


def test_get_page(mock_response):
    url = "https://www.scrapethissite.com/pages/simple/"

    with patch("requests.get") as mock_get:
        mock_get.return_value = mock_response
        page_content = ND_Open.get_page(url)

    assert page_content == "Hello World"


def test_get_html(mock_response):
    url = "https://example.com"

    with patch("requests.get") as mock_get:
        mock_get.return_value = mock_response
        soup = ND_Open.get_html(url)

    assert str(soup) == "<html><body><p>Hello World</p></body></html>"


def test_get_page_empty_url():
    with pytest.raises(ValueError):
        ND_Open.get_page("")


def test_get_html_empty_url():
    with pytest.raises(ValueError):
        ND_Open.get_html("")


def test_get_page_request_exception():
    with pytest.raises(Exception):
        ND_Open.get_page("")


def test_get_html_request_exception():
    with pytest.raises(Exception):
        ND_Open.get_html("")


def test_get_page_feature_not_found(mock_response):
    url = ""

    with patch("requests.get") as mock_get, patch("bs4.BeautifulSoup") as mock_bs:
        mock_get.return_value = mock_response
        mock_bs.side_effect = FeatureNotFound
        with pytest.raises(Exception):
            ND_Open.get_page(url)


def test_get_html_feature_not_found(mock_response):
    url = ""

    with patch("requests.get") as mock_get, patch("bs4.BeautifulSoup") as mock_bs:
        mock_get.return_value = mock_response
        mock_bs.side_effect = FeatureNotFound
        with pytest.raises(Exception):
            ND_Open.get_html(url)



"""
Tests for selenium commands._open
"""

from asrch.commands._open import get_page, get_html, get_image, get_js




def test_images():
    """
    Tests get image function
    """

    assert (
        "https://www.scrapethissite.com/static/images/scraper-icon.png"
        in get_image(
            "https://www.scrapethissite.com/pages/simple/", proxy=None, header=False
        )
    )


def test_text():
    """
    Tests get page funtion
    """

    assert "Countries of the World" in get_page(
        "https://www.scrapethissite.com/pages/simple/", proxy=None, header=False
    )


def test_html():
    """
    Tests HTML function
    """
    assert "<body>" in get_html(
        "https://www.scrapethissite.com/pages/simple/", proxy=None, header=False
    )


def test_js():
    """
    Tests javascript function
    """
    gen_iter = get_js(
        "https://www.scrapethissite.com/pages/simple/", proxy=None, header=False
    )
    try:
        next(gen_iter)
        assert "* Copyright" in next(gen_iter)
    except StopIteration:
        pass



from dataclasses import dataclass

from asrch.utils.constants import Colors

colors = Colors()


@dataclass
class Pet:
    age: int
    name: str
    health: int
    color: Colors

    @property
    def pet_name(self) -> str:
        return self.name

    @pet_name.setter
    def pet_name(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError("Error: Name must be a string")
        print(f"Name set: {value}")
        self.name = value


@dataclass
class Cat(Pet):
    character = "ᓚᘏᗢ"


@dataclass
class Dog(Pet):
    pass


@dataclass
class Bird(Pet):
    pass


@dataclass
class Monkey(Pet):
    character = "@('-')@"


my_cat = Cat(0, "", 0, Colors.NC)
my_cat.pet_name = "joe"



import toml


def highlight_word(word, text, color_code):
    return text.replace(word, f"\x1b[{color_code}m{word}\x1b[0m")


config = toml.load("undetected.toml")

# Basic section
url = config["basic"]["url"]

# Privacy section
proxy = config["privacy"]["proxy"]

# Commands settings
open_header = config["commands"]["open"]["header"]
open_pager = config["commands"]["open"]["pager"]
open_download = config["commands"]["open"]["download"]
open_log = config["commands"]["open"]["log"]
open_nodriver = config["commands"]["open"]["nodriver"]
open_parser = config["commands"]["open"]["parser"]

search_query = config["commands"]["search"]["query"]
search_header = config["commands"]["search"]["header"]
search_nodriver = config["commands"]["search"]["nodriver"]
search_log = config["commands"]["search"]["log"]

whois_header = config["commands"]["whois"]["header"]

find_element = config["commands"]["find"]["element"]
find_header = config["commands"]["find"]["header"]
find_log = config["commands"]["find"]["log"]
find_locator = config["commands"]["find"]["locator"]

quizlet_header = config["commands"]["quizlet"]["header"]
quizlet_pager = config["commands"]["quizlet"]["pager"]


settings_list: list = []
# Basic section
settings_list.append(("url", config["basic"]["url"]))

# Privacy section
settings_list.append(("proxy", config["privacy"]["proxy"]))

# Commands settings
settings_list.append(search_query)

print(search_query)

highlighted_settings_list = []
"""for variable, value in settings_list:
    if "open" in variable:
        variable = highlight_word("open", variable, "1;33")  
    elif "search" in variable:
        variable = highlight_word("search", variable, "1;34") 
    elif "whois" in variable:
        variable = highlight_word("whois", variable, "1;35")  
    elif "find" in variable:
        variable = highlight_word("find", variable, "1;36")  
    elif "quizlet" in variable:
        variable = highlight_word("quizlet", variable, "1;31") 
    highlighted_settings_list.append((variable, value))"""




import logging

from asrch.modules.logging_formatter import ColorFormatter

c_log = logging.getLogger(__name__)
c_log.setLevel(logging.DEBUG)

sh = logging.StreamHandler()

c_form = ColorFormatter("%(asctime)s|%(levelname)8s|%(message)s")

sh.setFormatter(c_form)

c_log.addHandler(sh)


def test_logs(caplog):
    c_log.debug("d_message")
    c_log.info("i_message")
    c_log.warning("w_message")
    c_log.error("e_message")
    c_log.critical("c_message")

    assert "test_logger_test.py" and "DEBUG" and "d_message" and "19" in caplog.text
    assert "test_logger_test.py" and "INFO" and "i_message" and "19" in caplog.text
    assert "test_logger_test.py" and "WARNING" and "w_message" and "19" in caplog.text
    assert "test_logger_test.py" and "ERROR" and "e_message" and "19" in caplog.text
    assert "test_logger_test.py" and "CRITICAL" and "c_message" and "19" in caplog.text

    caplog.clear()
